import os, tqdm
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy.wcs import WCS, NoConvergence
from astropy.table import Table
from astropy.nddata import Cutout2D
from astropy.utils.data import download_file
import requests
from bs4 import BeautifulSoup
import warnings
import urllib

from .mast import tic_by_contamination


def check_pointing(sector, camera, chip, path=None):
    """ Checks to see if a pointing model exists locally already.
    """
    # Tries to create a pointing model directory
    if path == None:
        pm_dir = '.'
    else:
        pm_dir = path

    searches = [
        's{0:04d}-{1}-{2}_tess_v2_pm.txt'.format(sector, camera, chip),
        'pointingModel_{0:04d}_{1}-{2}.txt'.format(sector, camera, chip),
    ]

    # Checks a directory of pointing models, if it exists
    # Returns the pointing model if it's in the pointing model directory
    for search in searches:
        if not os.path.isdir(pm_dir):
            continue
        pm_downloaded = os.listdir(pm_dir)
        pm = [i for i in pm_downloaded if search in i]
        if len(pm) > 0:
            return Table.read(os.path.join(pm_dir, pm[0]), format="ascii.basic")
    warnings.warn("couldn't find pointing model")


def load_pointing_model(pm_dir, sector, camera, chip):
    """ Loads in pointing model.
    """
    return check_pointing(sector, camera, chip, path=pm_dir)


def use_pointing_model(coords, pointing_model):
    """Applies pointing model to correct the position of star(s) on postcard.

    Parameters
    ----------
    coords : tuple
        (`x`, `y`) position of star(s).

    pointing_model : astropy.table.Table
        pointing_model for ONE cadence.

    Returns
    -------
    coords : tuple
        Corrected position of star(s).
    """
    pointing_model = np.reshape(list(pointing_model), (3,3))
    A = np.column_stack([coords[0], coords[1], np.ones_like(coords[0])])
    fhat = np.dot(A, pointing_model)
    return fhat


def pm_quality(time, sector, camera, chip, pm=None):
        """ Fits a line to the centroid motions using the pointing model.
            A quality flag is set if the centroid is > 2*sigma away from
                the majority of the centroids.
        """

        def outliers(x, y, poly, mask):
            dist = (y - poly[0]*x - poly[1])/np.sqrt(poly[0]**2+1**2)
            std  = np.std(dist)
            ind  = np.where((dist > 2*std) | (dist < -2*std))[0]
            mask[ind] = 1
            return mask

        cen_x, cen_y = 1024, 1024 # Uses a point in the center of the FFI
        cent_x,cent_y = [], []

        if pm is None:
            pm = load_pointing_model(pm_dir, sector, camera, chip)

        # Applies centroids
        for i in range(len(pm)):
            new_coords = use_pointing_model(np.array([cen_x, cen_y]), pm[i])
            cent_x.append(new_coords[0][0])
            cent_y.append(new_coords[0][1])
        cent_x = np.array(cent_x); cent_y = np.array(cent_y)

        # Finds gap in orbits
        t = np.diff(time)
        brk = np.where( t > np.mean(t)+2*np.std(t))[0][0]
        brk += 1

        # Initiates lists for each orbit
        x1 = cent_x[0:brk]; y1 = cent_y[0:brk]
        x2 = cent_x[brk:len(cent_x)+1];y2 = cent_y[brk:len(cent_y)+1]

        # Initiates masks
        mask1 = np.zeros(len(x1)); mask2 = np.zeros(len(x2))


        # Loops through and searches for points > 2 sigma away from distribution
        for i in np.arange(0,10,1):
            poly1  = np.polyfit(x1[mask1==0], y1[mask1==0], 1)
            poly2  = np.polyfit(x2[mask2==0], y2[mask2==0], 1)
            mask1  = outliers(x1, y1, poly1, mask1)
            mask2  = outliers(x2, y2, poly2, mask2)

        # Returns a total mask for each orbit
        return np.append(mask1, mask2)


def set_quality_flags(ffi_start, ffi_stop, shortCad_fn, sector, camera, chip,
                      pm=None):
    """ Uses the quality flags in a 2-minute target to create quality flags
        in the postcards.
    We create our own quality flag as well, using our pointing model.
    """
    # Obtains information for 2-minute target
    twoMin     = fits.open(shortCad_fn)
    twoMinTime = twoMin[1].data['TIME']-twoMin[1].data['TIMECORR']
    finite     = np.isfinite(twoMinTime)
    twoMinQual = twoMin[1].data['QUALITY']

    twoMinTime = twoMinTime[finite]
    twoMinQual = twoMinQual[finite]

    perFFIcad = []
    for i in range(len(ffi_start)):
        where = np.where( (twoMinTime > ffi_start[i]) &
                          (twoMinTime < ffi_stop[i]) )[0]
        perFFIcad.append(where)

    perFFIcad = np.array(perFFIcad)
    # Binary string for values which apply to the FFIs
    ffi_apply = int('100010101111', 2)

    convolve_ffi = []
    for cadences in perFFIcad:
        v = np.bitwise_or.reduce(twoMinQual[cadences])
        convolve_ffi.append(v)
    convolve_ffi = np.array(convolve_ffi)

    flags    = np.bitwise_and(convolve_ffi, ffi_apply)
    pm_flags = pm_quality(ffi_stop, sector, camera, chip, pm=pm) * 131072

    pm_flags[ ((ffi_stop>1420.) & (ffi_stop < 1424.)) ] = 131072

    return np.bitwise_or(flags, pm_flags)

def centroid_quadratic(data, mask=None):
    """Computes the quadratic estimate of the centroid in a 2d-array.

    This method will fit a simple 2D second-order polynomial
    $P(x, y) = a + bx + cy + dx^2 + exy + fy^2$
    to the 3x3 patch of pixels centered on the brightest pixel within
    the image.  This function approximates the core of the Point
    Spread Function (PSF) using a bivariate quadratic function, and returns
    the maximum (x, y) coordinate of the function using linear algebra.

    For the motivation and the details around this technique, please refer
    to Vakili, M., & Hogg, D. W. 2016, ArXiv, 1610.05873.

    Caveat: if the brightest pixel falls on the edge of the data array, the fit
    will tend to fail or be inaccurate.

    As used in the Lightkurve package of Barentsen et al.

    Parameters
    ----------
    data : 2D array
        The 2D input array representing the pixel values of the image.
    mask : array_like (bool), optional
        A boolean mask, with the same shape as `data`, where a **True** value
        indicates the corresponding element of data is masked.

    Returns
    -------
    column, row : tuple
        The coordinates of the centroid in column and row.  If the fit failed,
        then (NaN, NaN) will be returned.
    """
    # Step 1: identify the patch of 3x3 pixels (z_)
    # that is centered on the brightest pixel (xx, yy)
    if mask is not None:
        data = data * mask
    arg_data_max = np.nanargmax(data)
    yy, xx = np.unravel_index(arg_data_max, data.shape)
    # Make sure the 3x3 patch does not leave the TPF bounds
    if yy < 1:
        yy = 1
    if xx < 1:
        xx = 1
    if yy > (data.shape[0] - 2):
        yy = data.shape[0] - 2
    if xx > (data.shape[1] - 2):
        xx = data.shape[1] - 2

    z_ = data[yy-1:yy+2, xx-1:xx+2]

    # Next, we will fit the coefficients of the bivariate quadratic with the
    # help of a design matrix (A) as defined by Eqn 20 in Vakili & Hogg
    # (arxiv:1610.05873). The design matrix contains a
    # column of ones followed by pixel coordinates: x, y, x**2, xy, y**2.
    A = np.array([[1, -1, -1, 1,  1, 1],
                  [1,  0, -1, 0,  0, 1],
                  [1,  1, -1, 1, -1, 1],
                  [1, -1,  0, 1,  0, 0],
                  [1,  0,  0, 0,  0, 0],
                  [1,  1,  0, 1,  0, 0],
                  [1, -1,  1, 1, -1, 1],
                  [1,  0,  1, 0,  0, 1],
                  [1,  1,  1, 1,  1, 1]])
    # We also pre-compute $(A^t A)^-1 A^t$, cf. Eqn 21 in Vakili & Hogg.
    At = A.transpose()
    # In Python 3 this can become `Aprime = np.linalg.inv(At @ A) @ At`
    Aprime = np.matmul(np.linalg.inv(np.matmul(At, A)), At)

    # Step 2: fit the polynomial $P = a + bx + cy + dx^2 + exy + fy^2$
    # following Equation 21 in Vakili & Hogg.
    # In Python 3 this can become `Aprime @ z_.flatten()`
    a, b, c, d, e, f = np.matmul(Aprime, z_.flatten())

    # Step 3: analytically find the function maximum,
    # following https://en.wikipedia.org/wiki/Quadratic_function
    det = 4 * d * f - e ** 2
    if abs(det) < 1e-6:
        return np.nan, np.nan  # No solution
    xm = - (2 * f * b - c * e) / det
    ym = - (2 * d * c - b * e) / det
    return xx + xm, yy + ym


class ffi:
    """This class allows the user to download all full-frame images for a given sector,
         camera, and chip. It also allows the user to create their own pointing model
         based on each cadence for a given combination of sector, camera, and chip.

    No individual user should have to download all of the full-frame images because
         stacked postcards will be available for the user to download from MAST.

    Parameters
    ----------
    sector : int, optional
    camera : int, optional
    chip : int, optional
    """
    def __init__(self, sector=None, camera=None, chip=None):
        self.sector = sector
        self.camera = camera
        self.chip   = chip
        self.ffiindex = None

    def download_ffis(self, download_dir=None):
        """
        Downloads entire sector of data into FFI download directory.

        Parameters
        ----------
        download_dir : str
            Location where the data files will be stored.
            Defaults to "~/.eleanor/sector_{}/ffis" if `None` is passed.
        """

        def findAllFFIs():
            nonlocal url
            sub_paths = []
            subsub_paths = []
            calFiles, urlPaths = [], []

            paths = BeautifulSoup(requests.get(url).text, "lxml").find_all('a')

            for direct in paths:
                subdirect = direct.get('href')
                if ('2018' in subdirect) or ('2019' in subdirect):
                    sub_paths.append(os.path.join(url, subdirect))

            for sp in sub_paths:
                for fn in BeautifulSoup(requests.get(sp).text, "lxml").find_all('a'):
                    subsub = fn.get('href')
                    if (subsub[0] != '?') and (subsub[0] != '/'):
                        subsub_paths.append(os.path.join(sp, subsub))

            subsub_paths = [os.path.join(i, '{}-{}/'.format(self.camera, self.chip)) for i in subsub_paths]

            for sbp in subsub_paths:
                for fn in BeautifulSoup(requests.get(sbp).text, "lxml").find_all('a'):
                    if 'ffic.fits' in fn.get('href'):
                        calFiles.append(fn.get('href'))
                        urlPaths.append(sbp)

            return np.array(calFiles), np.array(urlPaths)

        # This URL applies to ETE-6 simulated data ONLY
        url = 'https://archive.stsci.edu/missions/tess/ffi/'

        url = os.path.join(url, "s{0:04d}".format(self.sector))
        files, urlPaths = findAllFFIs()

        if download_dir is None:
            # Creates hidden .eleanor FFI directory
            ffi_dir = self._fetch_ffi_dir()
        else:
            ffi_dir = download_dir

        files_in_dir = os.listdir(ffi_dir)

        local_paths = []
        for i in range(len(files)):
            if files[i] not in files_in_dir:
                os.system('cd {} && curl -O -L {}'.format(ffi_dir, urlPaths[i]+files[i]))
            local_paths.append(ffi_dir+files[i])
        self.local_paths = np.array(local_paths)
        return


    def _fetch_ffi_dir(self):
        """Returns the default path to the directory where FFIs will be downloaded.

        By default, this method will return "~/.eleanor/sector_{}/ffis" and create
        this directory if it does not exist.  If the directory cannot be
        access or created, then it returns the local directory (".").

        Returns
        -------
        download_dir : str
            Path to location of `ffi_dir` where FFIs will be downloaded
        """
        download_dir = os.path.join(os.path.expanduser('~'), '.eleanor',
                                    'sector_{}'.format(self.sector), 'ffis')
        if os.path.isdir(download_dir):
            return download_dir
        else:
            # if it doesn't exist, make a new cache directory
            try:
                os.makedirs(download_dir)
            # downloads locally if OS error occurs
            except OSError:
                warnings.warn('Warning: unable to create {}. '
                              'Downloading FFIs to the current '
                              'working directory instead.'.format(download_dir))
                download_dir = '.'

        return download_dir

    def sort_by_date(self):
        """Sorts FITS files by start date of observation."""
        dates, time, index = [], [], []
        for f in self.local_paths:
            hdu = fits.open(f)
            hdr = hdu[1].header
            dates.append(hdr['DATE-OBS'])
            if 'ffiindex' in hdu[0].header:
                index.append(hdu[0].header['ffiindex'])
        if len(index) == len(dates):
            dates, index, fns = np.sort(np.array([dates, index, self.local_paths]))
            self.ffiindex = index.astype(int)
        else:
            dates, fns = np.sort(np.array([dates, self.local_paths]))
        self.local_paths = fns
        self.dates = dates
        return


    def build_pointing_model(self, pos_predicted, pos_inferred, outlier_removal=False):
        """Builds an affine transformation to correct the positions of stars
           from a possibly incorrect WCS.

        Parameters
        ----------
        pos_predicted : tuple
            Positions taken straight from the WCS; [[x,y],[x,y],...] format.
        pos_inferred : tuple
            Positions taken using any centroiding method; [[x,y],[x,y],...] format.
        outlier_removal : bool, optional
            Whether to clip 1-sigma outlier frames. Default `False`.

        Returns
        -------
        xhat : np.ndarray
            (3, 3) affine transformation matrix between WCS positions
            and inferred positions.
        """
        A = np.column_stack([pos_predicted[:,0], pos_predicted[:,1], np.ones_like(pos_predicted[:,0])])
        f = np.column_stack([pos_inferred[:,0], pos_inferred[:,1], np.ones_like(pos_inferred[:,0])])
        if outlier_removal == True:
            dist = np.sqrt(np.sum((A - f)**2, axis=1))
            mean, std = np.nanmean(dist), np.nanstd(dist)
            lim = 1.0
            A = A[dist < mean + lim*std]
            f = f[dist < mean + lim*std]
        ATA = np.dot(A.T, A)
        ATAinv = np.linalg.inv(ATA)
        ATf = np.dot(A.T, f)
        xhat = np.dot(ATAinv, ATf)
        fhat = np.dot(A, xhat)

        return xhat


    def pointing_model_per_cadence(self, out_dir=None, n_sources=350):
        """Step through build_pointing_model for each cadence."""

        def find_isolated(x, y):
            """Finds the most isolated, least contaminated sources for pointing model."""
            init_d   = 8.0
            counter  = 0.0
            isolated = []

            while ((init_d - counter) > 0) and (counter < 3.5):
                for i in range(len(x)):
                    x_list  = np.delete(x, np.where(x==x[i]))
                    y_list  = np.delete(y, np.where(y==y[i]))
                    closest = np.sqrt( (x[i]-x_list)**2 + (y[i]-y_list)**2 ).argmin()
                    dist    = np.sqrt( (x[i]-x_list[closest])**2 + (y[i]-y_list[closest])**2 )
                    if (dist > (init_d - counter)) and (i not in isolated):
                        isolated.append(i)
                    if len(isolated) > n_sources:
                        break
                counter += 0.1

            return np.array(isolated)


        def isolated_center(x, y, image):
            """Finds the centroid of each isolated source with centroid_quadratic.

            Parameters
            ----------
            x : array-like
                Initial guesses of x positions for sources.
            y : array-like
                Initial guesses of y positions for sources.
            image : np.ndarray
                FFI flux data.

            Returns
            -------
            cenx : array-like
                Controid x coordinates for all good input sources.
            ceny : array-like
                Centroid y coordinates for all good input sources.
            good : list
                Indexes into input arrays corresponding to good sources.
            """
            cenx, ceny, good = [], [], []

            for i in range(len(x)):
                 if x[i] > 0. and y[i] > 0.:
                     tpf = Cutout2D(image, position=(x[i], y[i]), size=(7,7), mode='partial')
                     origin = tpf.origin_original
                     cen = centroid_quadratic(tpf.data - np.nanmedian(tpf.data))
                     cenx.append(cen[0]+origin[0]); ceny.append(cen[1]+origin[1])
                     good.append(i)
            cenx, ceny = np.array(cenx), np.array(ceny)
            return cenx, ceny, good

        def apply_pointing_model(xy, matrix):
            pointing_model = matrix
            centroid_xs, centroid_ys = [], []
            new_coords = use_pointing_model(np.array(xy).T, pointing_model)
            return np.array(new_coords)

        pm_fn = 's{0:04d}-{1}-{2}_tess_v2_pm.txt'.format(self.sector, self.camera, self.chip)

        eleanorpath = os.path.join(os.path.expanduser('~'), '.eleanor')

        qf = np.loadtxt(eleanorpath + '/metadata/s{0:04d}/quality_s{0:04d}.txt'.format(self.sector,
                                                                                                                        self.sector))

        if out_dir is not None:
            pm_fn = out_dir+ '/' + pm_fn

        with open(pm_fn, 'w') as tf:
            tf.write('0 1 2 3 4 5 6 7 8\n')

        pos = None
        for fn in self.local_paths:
            with fits.open(fn) as hdu:
                hdr = hdu[1].header
            if 'CRVAL1' not in hdr or 'CRVAL2' not in hdr:
                continue
            pos = [hdr['CRVAL1'], hdr['CRVAL2']]
        if pos is None:
            raise ValueError("no WCS")

        r = 6.0*np.sqrt(1.2)
        contam = [0.0, 0.5]
        tmag_lim = [7.5, 12.5]

        t  = tic_by_contamination(pos, r, contam, tmag_lim).group_by('contratio')

        for i, fn in enumerate(self.local_paths):
            with fits.open(fn) as hdu:
                hdr = hdu[1].header
                data = hdu[1].data

            try:
                xy = WCS(hdr).all_world2pix(t['ra'], t['dec'], 1)

                # Triple checks the sources are on the FFI
                onFrame = np.where( (xy[0]>10) & (xy[0]<2092-10) & (xy[1]>10) & (xy[1]<2048-10) )[0]
                xy  = np.array([xy[0][onFrame], xy[1][onFrame]])
                iso = find_isolated(xy[0], xy[1])
                if len(iso) > 0:
                    xy  = np.array([xy[0][iso], xy[1][iso]])
                    cenx, ceny, good = isolated_center(xy[0], xy[1], data)

                    # Triple checks there are no nans; Nans make people sad
                    no_nans = np.where( (np.isnan(cenx)==False) & (np.isnan(ceny)==False))
                    pos_inferred = np.array( [cenx[no_nans], ceny[no_nans]] )
                    xy = np.array( [xy[0][no_nans], xy[1][no_nans]] )

                    solution = self.build_pointing_model(xy.T, pos_inferred.T)

                    xy = apply_pointing_model(xy.T, solution)
                    matrix = self.build_pointing_model(xy, pos_inferred.T, outlier_removal=True)

                    sol    = np.dot(matrix, solution)
                    sol    = sol.flatten()
                else:
                    sol = np.full((9,), 1e5)

            except NoConvergence:
                if qf[i] != 0:
                    sol = np.full((9,), 1e5)
                else:
                    a   = np.zeros((3, 3), int)
                    np.fill_diagonal(a, 1)
                    sol = np.reshape(a, (9,))

            with open(pm_fn, 'a') as tf:
                tf.write('{}\n'.format(' '.join(str(e) for e in sol) ) )

        with open(pm_fn, "r") as tf:
            return Table.read(tf.read(), format='ascii.basic')



import os
from sherlockpipe.objectinfo.ObjectInfo import ObjectInfo

class InputObjectInfo(ObjectInfo):
    """
    Implementation of ObjectInfo to be used to characterize objects which are to be loaded from a csv file.
    """
    def __init__(self, input_file, initial_mask=None, initial_detrend_period=None, star_info=None,
                 aperture_file=None):
        """
        @param input_file: the file to be used for loading the light curve
        @param initial_mask: an array of time ranges provided to mask them into the initial object light curve.
        @param initial_detrend_period: integer value specifying a fixed value for an initial period to be detrended
        from the initial light curve before processing.
        @param star_info: input star information
        @param aperture_file: the file containing 1s and 0s specifying the user selected aperture
        """
        super().__init__(initial_mask, initial_detrend_period, star_info, aperture_file)
        self.input_file = input_file

    def sherlock_id(self):
        return "INP_" + os.path.splitext(self.input_file)[0].replace("/", "_")

    def mission_id(self):
        return None



__all__ = ['EleanorWarning', 'SearchError']

class EleanorWarning(Warning):
    """A class to hold Eleanor-specific warnings."""
    pass

class SearchError(Exception):
    """Exception raised when no target was found."""
    pass



from contextlib import contextmanager
from timeit import default_timer
from sherlockpipe.sherlock import Sherlock
from sherlockpipe.objectinfo.InputObjectInfo import InputObjectInfo
from sherlockpipe.objectinfo.MissionFfiCoordsObjectInfo import MissionFfiCoordsObjectInfo
from sherlockpipe.objectinfo.MissionFfiIdObjectInfo import MissionFfiIdObjectInfo
from sherlockpipe.objectinfo.MissionInputObjectInfo import MissionInputObjectInfo
from sherlockpipe.objectinfo.MissionObjectInfo import MissionObjectInfo


@contextmanager
def elapsed_timer():
    start = default_timer()
    elapser = lambda: str(default_timer() - start)
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: str(end - start)


with elapsed_timer() as elapsed:
    # Adding several kinds of objects to the run: one short cadence TIC, one FFI TIC, one coordinates FFI, one input
    # file related to a TIC and one plain input file.
    # Ensure that your input light curve CSV files have three columns: #TBJD,flux,flux_err
    sherlock = Sherlock([MissionObjectInfo("TIC 181804752", 'all'),
                                        MissionFfiIdObjectInfo("TIC 259168516", [14, 15]),
                                        MissionFfiCoordsObjectInfo(14, 19, 'all'),
                                        MissionInputObjectInfo("TIC 470381900", "example_lightcurve.csv"),
                                        InputObjectInfo("example_lc.csv")])\
        .run()
    print("Analysis took " + elapsed() + "s")



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()
version = os.popen('git tag -l --sort -version:refname | head -n 1').read().split('\n', 1)[0]

setuptools.setup(
    name="sherlockpipe", # Replace with your own username
    version=version,
    author="F.J. Pozuelos & M. Dévora",
    author_email="fjpozuelos@uliege.be",
    description="Search for Hints of Exoplanets fRom Lightcurves Of spaCe based seeKers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/franpoz/SHERLOCK",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["numpy",
                        "cython",
                        "pandas",
                        "lightkurve",
                        "transitleastsquares",
                        "requests",
                        "wotan",
                        "matplotlib",
                        "pyyaml",
                        "allesfitter",
                        "dynesty",
                        "emcee",
                        "corner",
                        "ellc",
                        "seaborn",
                        "bokeh",
                        "astroplan",
                        "astroquery",
                        "sklearn",
                        "scipy",
                        "tess-point",
                        "reproject==0.4",
                        "reportlab",
                        "astropy",
                        "mock > 2.0.0",
                        'photutils>=0.7',
                        'tqdm',
                        'astropy>=3.2.3',
                        'setuptools>=41.0.0',
                        'torch',
                        'beautifulsoup4>=4.6.0',
                        'tess-point>=0.3.6'
    ]
)



import numpy as np

from sherlockpipe.scoring.BasicSignalSelector import BasicSignalSelector
from sherlockpipe.scoring.SnrBorderCorrectedSignalSelector import CorrectedBorderSignalSelection


class QuorumSnrBorderCorrectedSignalSelector(BasicSignalSelector):
    def __init__(self, strength=1, min_quorum=0):
        super().__init__()
        self.strength = strength
        self.min_quorum = min_quorum

    def select(self, transit_results, snr_min, detrend_method, wl):
        basic_signal_selection = super().select(transit_results, snr_min, detrend_method, wl)
        index_snr_period_t0_array = [[key, transit_result.snr * transit_result.border_score,
                                        transit_result.period, transit_result.t0]
                                        for key, transit_result in transit_results.items()]
        # index_snr_period_t0_array = sorted(index_snr_period_t0_array, key=lambda k: k[2])
        i = 0
        number_of_voters = len(index_snr_period_t0_array)
        votes = list(range(0, number_of_voters))
        while i < len(index_snr_period_t0_array):
            if i != votes[i]:  # we already compared the vote
                i = i + 1
                continue
            period = index_snr_period_t0_array[i][2]
            t0 = index_snr_period_t0_array[i][3]
            j = i + 1
            while j < len(index_snr_period_t0_array):
                compared_period = index_snr_period_t0_array[j][2]
                compared_t0 = index_snr_period_t0_array[j][3]
                if abs(period - compared_period) < 0.025 and abs(t0 - compared_t0) < 0.01:
                    votes[j] = i
                j = j + 1
            i = i + 1
        votes_counts = [votes.count(vote) for vote in votes]
        corrected_snrs = [index_snr_period_t0_array[key][1] +
                          index_snr_period_t0_array[key][1] * (votes_count - 1) / number_of_voters * self.strength
                for key, votes_count in enumerate(votes_counts)]
        best_signal_snr = np.nanmax(corrected_snrs)
        best_signal_snr_index = np.nanargmax(corrected_snrs)
        best_signal = transit_results[best_signal_snr_index]
        max_votes_rate = max(votes_counts) / len(votes)
        if best_signal_snr > snr_min and max_votes_rate >= self.min_quorum:  # and SDE[a] > SDE_min and FAP[a] < FAP_max):
            best_signal_score = 1
        else:
            best_signal_score = 0
        return CorrectedQuorumBorderSignalSelection(best_signal_score, best_signal_snr,
                                                    basic_signal_selection.curve_index,
                                                    transit_results[basic_signal_selection.curve_index],
                                                    best_signal_snr_index, best_signal,
                                                    votes_counts[best_signal_snr_index])

class CorrectedQuorumBorderSignalSelection(CorrectedBorderSignalSelection):
    def __init__(self, score, corrected_snr, original_curve_index, original_transit_result, final_curve_index,
                 final_transit_result, votes):
        super().__init__(score, corrected_snr, original_curve_index, original_transit_result, final_curve_index,
                         final_transit_result)
        self.votes = votes

    def get_message(self):
        curve_name = "PDCSAP_FLUX" if self.curve_index == 0 else str(self.curve_index - 1)
        original_curve_name = "PDCSAP_FLUX" if self.original_curve_index == 0 else str(self.original_curve_index - 1)
        return "Elected signal with QUORUM algorithm from " + str(self.votes) + " VOTES --> NAME: " + curve_name + \
               "\tPeriod:" + str(self.transit_result.period) + \
               "\tCORR_SNR: " + str(self.corrected_snr) + \
               "\tSNR: " + str(self.transit_result.snr) + \
               "\tSDE: " + str(self.transit_result.sde) + \
               "\tFAP: " + str(self.transit_result.fap) + \
               "\tBORDER_SCORE: " + str(self.transit_result.border_score) + \
               "\nProposed selection with BASIC algorithm was --> NAME: " + original_curve_name + \
               "\tPeriod:" + str(self.original_transit_result.period) + \
               "\tSNR: " + str(self.original_transit_result.snr)



import numpy as np
from sherlockpipe.star.StarCatalog import StarCatalog
from astroquery.mast import Catalogs
import transitleastsquares as tls


class TicStarCatalog(StarCatalog):
    def __init__(self):
        super().__init__()

    def catalog_info(self, id):
        """Takes TIC_ID, returns stellar information from online catalog using Vizier"""
        if type(id) is not int:
            raise TypeError('TIC_ID ID must be of type "int"')
        result = Catalogs.query_criteria(catalog="Tic", ID=id).as_array()
        Teff = result[0][64]
        lum = result[0]['lum']
        logg = result[0][66]
        radius = result[0][70]
        ra = result[0][13]
        dec = result[0][14]
        radius_max = result[0][71]
        radius_min = result[0][71]
        mass = result[0][72]
        mass_max = result[0][73]
        mass_min = result[0][73]
        if lum is None or np.isnan(lum):
            lum = self.star_luminosity(Teff, radius)
        ld, mass, mass_min, mass_max, radius, radius_min, radius_max = tls.catalog_info(TIC_ID=id)
        return (ld, Teff, lum, logg, radius, radius_min, radius_max, mass, mass_min, mass_max, ra, dec)



from contextlib import contextmanager
from timeit import default_timer
from sherlockpipe.sherlock import Sherlock
from sherlockpipe.objectinfo.InputObjectInfo import InputObjectInfo
from sherlockpipe.objectinfo.MissionFfiIdObjectInfo import MissionFfiIdObjectInfo
from sherlockpipe.objectinfo.MissionInputObjectInfo import MissionInputObjectInfo
from sherlockpipe.objectinfo.MissionObjectInfo import MissionObjectInfo


@contextmanager
def elapsed_timer():
    start = default_timer()
    elapser = lambda: str(default_timer() - start)
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: str(end - start)


with elapsed_timer() as elapsed:
    # We will use one FFI object from TESS, one short cadence object from TESS restricted to one sector, one short
    # cadence object from Kepler and one short cadence object from K2.
    # We will also provide two objects whose light curve info source are input files. For the first one, we will mask
    # two time intervals and for the second one we will add an initial detrend period.
    # We will:
    # 1 Enable the initial smooth function, which reduces local noise in the signal.
    # 2 Enable the initial High RMS areas masking. This procedure will mask all the lightcurve time binned ranges by
    # the 'initial_rms_bin_hours' value with a threshold of the 'initial_rms_threshold' value * RMS_median.
    # 3 Set the number of detrends to be done from PDCSAP_FLUX for each run.
    # 4 Set the SHERLOCK PDCSAP_FLUX detrends method to Gaussian Processes.
    # 5 Set the number of CPU cores to be used by the detrending procedure.
    # 6 Enable the Auto-Detrend detection, which will search for strong periodicities in the light curve and do an
    # initial detrend for it based on the selected 'auto_detrend_method' method and the value of the
    # 'auto_detrend_ratio' value, which ensures that we are detrending the light curve at 'auto_detrend_ratio' times
    # the stronger period.

    # 7 Set the maximum number of runs to be executed.
    # 8 Select the period protect value, which TODO
    # 9 Select the min period for a transit to be fit.
    # 10 Select the max period for a transit to be fit.
    # 11 Select the binning for TODO
    # 12 Select the number of CPU cores to be used for the transit search.
    # 13 Select the min SNR, the min SDE and the max FAP to stop the runs execution for each object.
    # 14 Select the found transits masking method. We use subtract here as example, but it is discouraged.
    # 15 Select the best signal algorithm, which provides a different implementation to decide which of the detrend
    # signals is the stronger one to be selected.
    # 10 Set the strength of the quorum algorithm votes, which makes every vote that is found to increase the SNR by
    # a factor of 1.2 for our selection.
    sherlock = Sherlock([MissionFfiIdObjectInfo("TIC 181804752", 'all'),
                                        MissionObjectInfo("TIC 259168516", [15]),
                                        MissionObjectInfo('KIC 10905746', 'all'),
                                        MissionObjectInfo('EPIC 249631677', 'all'),
                                        MissionInputObjectInfo("TIC 181804752", 'example_lc.csv',
                                                               initial_mask=[[1625, 1626], [1645, 1646]]),
                                        InputObjectInfo("example_lc.csv", initial_detrend_period=0.8)]) \
        .setup_detrend(initial_smooth=True, initial_rms_mask=True, initial_rms_threshold=2.5, initial_rms_bin_hours=3,
                       n_detrends=12, detrend_method="gp", cores=2, auto_detrend_periodic_signals=True,
                       auto_detrend_ratio=1 / 3, auto_detrend_method="cosine") \
        .setup_transit_adjust_params(max_runs=10, period_protec=12, period_min=1, period_max=10, bin_minutes=20,
                                 run_cores=3, snr_min=6, sde_min=6, fap_max=0.08, mask_mode="subtract",
                                 best_signal_algorithm='quorum', quorum_strength=1.2)\
        .run()
    print("Analysis took " + elapsed() + "s")



###
### The SHERLOCK PIPEline; Searching for Hints of Exoplanets fRom Lightcurves Of spaCe-based seeKers 
### 
### Pipeline to download, detrend and search for transits in 2-min cadence TESS data
###
### Pozuelos, Thuillier & Garcia 
### V13 10/06/2020
###
### Change Log:
###             
###               This version chooses the signal with lartest SNR to keep searching. The SDE and FAP are printed but   
###               are not used to chose which is the best signal. This is motivated by a bug discovered by Max 
###               It has been reported to TLS developers, it will be fix soon. 
###
###


import os, sys
from contextlib import contextmanager

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from wotan import flatten
from wotan import t14
import warnings
from astropy.stats import sigma_clip
from transitleastsquares import transitleastsquares
from transitleastsquares import transit_mask
from transitleastsquares import catalog_info
from transitleastsquares import cleaned_array
import lightkurve as lk
from lightkurve import search_lightcurvefile
from scipy import stats

# -------------------------------------------------------------------------------------------------------------------------------------------------------------
# User definitions
# -------------------------------------------------------------------------------------------------------------------------------------------------------------

TIC_ID = 254113311    # TIC_ID of our candidate
det_met = 1          # detrended method: =1 for bi-weigth ; =2 for GP matern 3/2
N_detrends = 12.      # Num. of detrend models applied
time_units = 1       # units of time: =0 (julian day) ; =1 (barycenter tess julian day)
P_protec = 10        # We procted the transit for a hypothetical Earth-size planet of Period = P_protec days. By default set = 1
Pmin = 0.5           # Minimun period in units of day
Pmax = 20            # Maximum period in units of day
minutes = 10         # time in minutes to bin the lightcurve
SNR_min = 5          # Threshold limit to keep searching. Strong restriction =>10
SDE_min = 5          # Threshold limit to keep searching. Strong restriction =>15
FAP_max = 1e-1       # Threshold limit to keep searching. Strong restriction <=1e-4
mask = 0             # = 1 means you need a mask; =0 means you do not want to use a mask 
Version = 13

# No need to change anything else from this point to down.


# *************************************************************************************************************************************************************
# Function that prints texts in a .log file
# *************************************************************************************************************************************************************
def logprint(*text):
#    print(*text)
    original = sys.stdout
    with open( os.path.join('TIC'+str(TIC_ID)+'_report.log'), 'a' ) as f:
        sys.stdout = f
        print(*text)
    sys.stdout = original


# *************************************************************************************************************************************************************
# Function that trim the axis (remove 'N' points)
# *************************************************************************************************************************************************************
def trim_axs(axs, N):
    axs = axs.flat
    for ax in axs[N:]:
        ax.remove()
    return axs[:N]


# *************************************************************************************************************************************************************
# Function that compute the data
# *************************************************************************************************************************************************************
def analyse(det_met, time_i, flatten_i, n_tra, ab, mass, mass_min, mass_max, radius, radius_min, radius_max, id_run) :
    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    # Detrending of the data using WOTAN
    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    # By default is used a 'biweight' method.
    # It will detrend the lightcurve 'N_detrends' independent times
    from wotan import flatten

    if det_met == 2 :
        wl_min = 1
        wl_max = 12
    else :
        tdur = t14(R_s=radius, M_s=mass, P=P_protec, small_planet=True) # we define the typical duration of a small planet in this star
        wl_min = 3*tdur # minimum transit duration
        wl_max = 20*tdur # maximum transit duration

    wl_step = (wl_max-wl_min)/N_detrends
    wl = np.arange(wl_min,wl_max,wl_step) # we define all the posibles window_length that we apply
    global_flatten_lc = np.zeros((len(wl),len(flatten_i)))
    global_trend_lc = np.zeros((len(wl),len(flatten_i)))

    for i in range(0,len(wl)):
        if det_met == 2 :
            flatten_lc, trend_lc  = flatten(time_i, flatten_i, method='gp', kernel='matern', kernel_size=wl[i], return_trend=True, break_tolerance=0.5)
        else :
            flatten_lc, trend_lc  = flatten(time_i, flatten_i, window_length=wl[i], return_trend=True, method='biweight', break_tolerance=0.5 )
        global_flatten_lc[i] = flatten_lc
        global_trend_lc[i] = trend_lc

    ## save in the log file all the information concerning the detrendins applied
    warnings.filterwarnings("ignore")

    global_final = np.zeros((len(wl),len(flatten_i)))

    logprint('\n MODELS IN THE DETRENDING - Run '+str(id_run))
    logprint('========================================\n')

    if det_met == 2 :
        det_model = 'kernel_size:'
    else :
        det_model = 'window_size:'

    bins = len(time_i) * 2 / minutes
    bin_means, bin_edges, binnumber = stats.binned_statistic(time_i, flatten_i, statistic='mean', bins=bins)
    bin_stds, _, _ = stats.binned_statistic(time_i, flatten_i, statistic='std', bins=bins)
    bin_width = (bin_edges[1] - bin_edges[0])
    bin_centers = bin_edges[1:] - bin_width/2
    
    logprint('PDCSAP_FLUX_'+str(id_run),'\t ',' ------ ' , '\t', '------', '\t', '\t','RMS (ppm):', np.std(flatten_i)*1e6, '\t','RMS_10min (ppm):', np.std(bin_means[~np.isnan(bin_means)])*1e6 )
    for i in range(len(wl)):
        flatten = sigma_clip(global_flatten_lc[i], sigma_lower=20, sigma_upper=3)
        global_final[i] = flatten
        bins_i=len(time_i)*2/minutes
        bin_means_i, bin_edges_i, binnumber_i = stats.binned_statistic(time_i, flatten, statistic='mean', bins=bins_i)
        bin_stds_i, _, _ = stats.binned_statistic(time_i, flatten, statistic='std', bins=bins_i)
        bin_width_i = (bin_edges_i[1] - bin_edges_i[0])
        bin_centers_i = bin_edges_i[1:] - bin_width_i/2
        logprint('flatten_lc%s' %i ,'\t ','trend_lc%s' %i, '\t', det_model, format(wl[i],'0.4f'), '\t', 'RMS (ppm):', np.std(flatten)*1e6,'\t', 'RMS_10min (ppm):', np.std(bin_means_i[~np.isnan(bin_means_i)])*1e6 )

    ## save in a plot all the detrendings and all the data to inspect visually.
    cases = np.zeros((len(wl),1))
    for i in range(len(wl)):
        cases[i] = wl[i]

    figsize = (8, 8) #x,y
    cols = 3
    rows = len(cases) // cols

    shift = 2*(1.0-(np.min(flatten_i))) #shift in the between the raw and detrended data
    ylim_max = 1.0+3*(np.max(flatten_i)-1.0) #shift in the between the raw and detrended data
    ylim_min = 1.0-2.0*shift #shift in the between the raw and detrended data

    fig1, axs = plt.subplots(rows, cols, figsize=figsize, constrained_layout=True)
    axs = trim_axs(axs, len(cases))
    for ax, case in zip(axs, cases):
        if det_met == 2 :
            ax.set_title('ks=%s' % str(np.around(case, decimals=4)))
        else:
            ax.set_title('ws=%s' % str(np.around(case, decimals=4)))

        bins_i=len(time_i)*2/minutes
        bin_means_i, bin_edges_i, binnumber_i = stats.binned_statistic(time_i, global_final[np.nonzero(cases == case)[0][0]], statistic='mean', bins=bins_i)
        bin_stds_i, _, _ = stats.binned_statistic(time_i, global_final[np.nonzero(cases == case)[0][0]], statistic='std', bins=bins_i)
        bin_width_i = (bin_edges_i[1] - bin_edges_i[0])
        bin_centers_i = bin_edges_i[1:] - bin_width_i/2
        
        ax.plot(time_i, flatten_i, linewidth=0.05 ,color='black', alpha=0.75, rasterized=True)
        ax.plot(time_i, global_trend_lc[np.nonzero(cases == case)[0][0]], linewidth=1, color='orange', alpha=1.0)
        ax.plot(time_i, global_final[np.nonzero(cases == case)[0][0]]-shift, linewidth=0.05 ,color='teal', alpha=0.75, rasterized=True)
        ax.plot(bin_centers_i, bin_means_i-shift,  marker='.', markersize=2, color='firebrick', alpha=0.5, linestyle='none', rasterized=True)
        
        ax.set_ylim(ylim_min, ylim_max)
        plt.savefig('Detrends_'+'run_'+str(id_run)+'_TIC'+str(TIC_ID)+'.png', dpi=200)

    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    # Search of signals in the raw data
    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    logprint('\n SEARCH OF SIGNALS - Run', id_run)
    logprint('=================================\n')

    logprint('PDCSAP_FLUX_'+str(id_run), '\t', 'Period' ,'\t', 'Per_err', '\t', 'N.Tran', '\t', 'Mean Depth(ppt)', '\t', 'T. dur(min)', '\t','T0', '\t', 'SNR', '\t', 'SDE','\t', 'FAP\n')
    model = transitleastsquares(time_i, flatten_i)
    results_pdcsap = model.power(u=ab, M_star=mass, M_star_min=mass_min, M_star_max=mass_max, R_star=radius, R_star_min=radius_min, R_star_max=radius_max, period_min=Pmin, period_max=Pmax, n_transits_min=n_tra, show_progress_bar=False)
    x=[]

    if results_pdcsap.T0 != 0:
        for j in range(0,len(results_pdcsap.transit_depths)):
            # print (results.transit_depths[i])
            x = np.append(x, results_pdcsap.transit_depths[j])
            x = x[~np.isnan(x)]
            depth = (1.-np.mean(x))*100/0.1 #we change to ppt units
    else:
        depth = results_pdcsap.transit_depths

    logprint('-----', '\t ',
        format(results_pdcsap.period, '.5f'),'\t ',
        format(results_pdcsap.period_uncertainty, '.6f'),'\t ',
        results_pdcsap.distinct_transit_count,'\t',
        format(depth,'.3f'), '\t',
        format(results_pdcsap.duration*24*60, '.1f'),'\t',
        results_pdcsap.T0,'\t ',
        format(results_pdcsap.snr,'.3f'),'\t ',
        format(results_pdcsap.SDE,'.3f'), '\t ',
        results_pdcsap.FAP)
    logprint('\n')

    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    # Saving a plot of the best signal from raw data
    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    fig = save_plot(time_i, global_final, results_pdcsap, i)

    fig.suptitle('PDCSAP_FLUX_'+str(id_run)+' ## SNR:'+
        str(format(results_pdcsap.snr,'.3f'))+' ## SDE:'+
        str(format(results_pdcsap.SDE,'.3f'))+' ## FAP:'+
        str(results_pdcsap.FAP))
    plt.savefig('Run_'+str(id_run)+'_PDCSAP-FLUX_'+'TIC'+str(TIC_ID)+'.png', bbox_inches='tight', dpi=200)

    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    # Search of signals in the detrended data
    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    if det_met ==2 :
        logprint('Kernel_size','\t', 'Period' ,'\t', 'Per_err','\t', 'N.Tran','\t','Mean Depth(ppt)','\t','T. dur(min)','\t','T0', '\t', 'SNR', '\t', 'SDE','\t', 'FAP\n')
    else:
        logprint('Window_size','\t', 'Period' ,'\t', 'Per_err','\t', 'N.Tran','\t','Mean Depth(ppt)','\t','T. dur(min)','\t','T0', '\t', 'SNR', '\t', 'SDE','\t', 'FAP\n')
    SNR = np.zeros((len(wl), 1))
    SDE = np.zeros((len(wl), 1))
    FAP = np.zeros((len(wl), 1))
    periods = np.zeros((len(wl), 1))
    per_err = np.zeros((len(wl), 1))
    durations = np.zeros((len(wl), 1))
    tos = np.zeros((len(wl), 1))

    for i in range(len(wl)):
        print(i)
        model = transitleastsquares(time_i, global_final[i])
        results = model.power(u=ab, M_star=mass, M_star_min=mass_min, M_star_max=mass_max, R_star=radius, R_star_min=radius_min, R_star_max=radius_max, period_min=Pmin, period_max=Pmax, n_transits_min=n_tra, show_progress_bar=False)
        SNR[i] = results.snr
        SDE[i] = results.SDE
        FAP[i] = results.FAP
        periods[i] = results.period
        per_err[i] = results.period_uncertainty
        durations[i] = results.duration
        tos[i] = results.T0
        x = []
        if results.T0 != 0:
            for j in range(0, len(results.transit_depths)):
                # print (results.transit_depths[i])
                x=np.append(x, results.transit_depths[j])
                x = x[~np.isnan(x)]
                depth = (1.-np.mean(x))*100/0.1 #we change to ppt units
        else:
            depth = results.transit_depths
        logprint(format(wl[i], '.4f'), '\t ',
            format(results.period, '.5f'),'\t ',
            format(results.period_uncertainty, '.6f'),'\t ',
            results.distinct_transit_count,'\t',
            format(depth,'.3f'), '\t',
            format(results.duration*24*60, '.1f'),'\t',
            results.T0,'\t ',
            format(results.snr,'.3f'),'\t ',
            format(results.SDE,'.3f'), '\t ',
            results.FAP)

        # -----------------------------------------------------------------------------------------------------------------------------------------------------
        # Saving a plot of the best signal from detrended data
        # -----------------------------------------------------------------------------------------------------------------------------------------------------
        fig = save_plot(time_i, global_final, results, i)

        if det_met ==2 :
            fig.suptitle('Run '+str(id_run)+' ## kernel_size:'+str(format(wl[i], '.4f'))+' ## SNR:'+str(format(results.snr,'.3f'))+' ## SDE:'+str(format(results.SDE,'.3f'))+' ## FAP:'+str(results.FAP))
            #plt.savefig('TIC'+str(TIC_ID)+'_ks='+str(format(wl[i],'.4f'))+'_run_'+str(id_run)+'.png', bbox_inches='tight', dpi=200)
            plt.savefig('Run_'+str(id_run)+'_ks='+str(format(wl[i],'.4f'))+'_TIC'+str(TIC_ID)+'.png', bbox_inches='tight', dpi=200)
        else:
            fig.suptitle('Run '+str(id_run)+' ## window_size:'+str(format(wl[i], '.4f'))+' ## SNR:'+str(format(results.snr,'.3f'))+' ## SDE:'+str(format(results.SDE,'.3f'))+' ## FAP:'+str(results.FAP))
            #plt.savefig('TIC'+str(TIC_ID)+'_ws='+str(format(wl[i],'.4f'))+'_run_'+str(id_run)+'.png', bbox_inches='tight', dpi=200)
            plt.savefig('Run_'+str(id_run)+'_ws='+str(format(wl[i],'.4f'))+'_TIC'+str(TIC_ID)+'.png', bbox_inches='tight', dpi=200)
    SNR=np.nan_to_num(SNR)
    a = np.nanargmax(SNR)  #check the maximum SRN

    if results_pdcsap.snr > SNR[a]:
        logprint('\nBest Signal -->', '\t','PDCSAP_FLUX' ,'\t', 'SNR:',format(results_pdcsap.snr,'.3f'))

        if (results_pdcsap.snr > SNR_min): # and results_pdcsap.SDE > SDE_min and results_pdcsap.FAP < FAP_max):
            logprint('\nBest Signal is good enough to keep searching. Going to the next run.')
            key=1
        else:
            logprint('\nBest Signal does not look very promising. End')
            key=0
    else:
        if det_met ==2:
            logprint('\nBest Signal -->', '\t','flatten_lc'+str(a) ,'\t', 'kernel_size:', format(wl[a],'0.4f'),'\t' ,'SNR:',format(SNR[a][0],'.3f'))
        else:
            logprint('\nBest Signal -->', '\t','flatten_lc'+str(a),'\t' , 'window_size:', format(wl[a],'0.4f'),'\t' ,'SNR:',format(SNR[a][0],'.3f'))

        if (SNR[a] > SNR_min): #and SDE[a] > SDE_min and FAP[a] < FAP_max):
            logprint('\nBest Signal is good enough to keep searching. Going to the next run.')
            key=1
        else:
            logprint('\nBest Signal does not look very promising. End')
            key=0

    print("### SNR :", str(format(results.snr,'.3f')), "   SDE :", str(format(results.SDE,'.3f')), "   FAP :", str(results.FAP))

    return results_pdcsap, SNR, key, periods, durations, tos


# *************************************************************************************************************************************************************
# Function that save a plot of the data
# *************************************************************************************************************************************************************
def save_plot(time_i, global_final, results, i) :
    #start the plotting
    fig, (ax1,ax2,ax3) = plt.subplots(nrows=1, ncols=3, figsize=(10,3), constrained_layout=True)
    #1-Plot all the transits
    in_transit = transit_mask(time_i, results.period, results.duration, results.T0)
    ax1.scatter(time_i[in_transit], global_final[i][in_transit], color='red', s=2, zorder=0)
    ax1.scatter(time_i[~in_transit], global_final[i][~in_transit], color='black', alpha=0.05, s=2, zorder=0)
    ax1.plot(results.model_lightcurve_time, results.model_lightcurve_model, alpha=1, color='red', zorder=1)
    #plt.scatter(time_n, flux_new_n, color='orange', alpha=0.3, s=20, zorder=3)
    plt.xlim(time_i.min(), time_i.max())
    #plt.xlim(1362.0,1364.0)
    ax1.set(xlabel='Time (days)', ylabel='Relative flux')

    #phase folded plus binning
    ax2.plot(results.model_folded_phase, results.model_folded_model, color='red')
    ax2.scatter(results.folded_phase, results.folded_y, color='black', s=10, alpha=0.05, zorder=2)
    if (results.SDE) != 0:
        bins = 200
        bin_means, bin_edges, binnumber = stats.binned_statistic(results.folded_phase, results.folded_y, statistic='mean', bins=bins)
        bin_stds, _, _ = stats.binned_statistic(results.folded_phase, results.folded_y, statistic='std', bins=bins)
        bin_width = (bin_edges[1] - bin_edges[0])
        bin_centers = bin_edges[1:] - bin_width/2
        ax2.errorbar(bin_centers, bin_means, yerr=bin_stds/2, xerr=bin_width/2, marker='o', markersize=4, color='teal', alpha=1, linestyle='none')
        ax2.set_xlim(0.45, 0.55)
        ax2.set(xlabel='Phase', ylabel='Relative flux')
        plt.ticklabel_format(useOffset=False)
    else:
        ax2.set_xlim(0.45, 0.55)
        ax2.set(xlabel='Phase', ylabel='Relative flux')
        plt.ticklabel_format(useOffset=False)

    #SDE
    ax3 = plt.gca()
    ax3.axvline(results.period, alpha=0.4, lw=3)
    plt.xlim(np.min(results.periods), np.max(results.periods))
    for n in range(2, 10):
        ax3.axvline(n*results.period, alpha=0.4, lw=1, linestyle="dashed")
        ax3.axvline(results.period / n, alpha=0.4, lw=1, linestyle="dashed")
    ax3.set(xlabel='Period (days)', ylabel='SDE')
    ax3.plot(results.periods, results.power, color='black', lw=0.5)
    ax3.set_xlim(0., max(results.periods))

    return fig


# *************************************************************************************************************************************************************
# Function that add a mask on the raw data
# *************************************************************************************************************************************************************
def add_mask(time_i, flatten_i, results_pdcsap, periods, durations, SNR, tos, a):
    if results_pdcsap.snr > SNR[a]:
        intransit = transit_mask(time_i, results_pdcsap.period, 2*results_pdcsap.duration, results_pdcsap.T0)
    else:
        intransit = transit_mask(time_i, periods[a], 2*durations[a], tos[a])

    flatten_i[intransit]=np.nan
    time_j, flatten_j = cleaned_array(time_i, flatten_i)

    return time_j, flatten_j


def run():
    # *************************************************************************************************************************************************************
    # Main
    # *************************************************************************************************************************************************************

    # -------------------------------------------------------------------------------------------------------------------------------------------------------------
    #      Initialization
    # -------------------------------------------------------------------------------------------------------------------------------------------------------------

    # Search of data by lightkurve function concerning the TIC_ID provided
    # It retrieves all the data (sectors) available.
    # save in log file the info
    ab, mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(TIC_ID=TIC_ID)
    #search and donwload the TESS data using lightkurve
    lcf= lk.search_lightcurvefile('TIC '+str(TIC_ID), mission="tess").download_all()

    #we set the minum number of transits based on the sectors available
    if len(lcf) > 1:
        n_tra = 2
    else:
        n_tra = 1

    # saving usere definitions to keep the track of what have be done already
    logprint('\n SHERLOCK (Searching for Hints of Exoplanets fRom Lightcurves Of spaCe-base seeKers)')
    logprint('\n Version ', Version )
    logprint('\n USER DEFINITIONS')
    logprint('========================\n')
    logprint('TIC_ID: ', TIC_ID)

    if det_met == 2 :
        logprint('Detrend method: Gaussian Process Matern 2/3')
    else:
        logprint('Detrend method: Bi-Weight')

    logprint('No of detrend models applied: ', N_detrends)
    logprint('No of sectors available: ', len(lcf))
    logprint('Minimum number of transits: ', n_tra)
    logprint('Period planet protected: ', P_protec)
    logprint('Minimum Period (d): ', Pmin)
    logprint('Maximum Period (d): ', Pmax)
    logprint('Binning size (min): ', minutes)

    if time_units ==1 :
        logprint('Units of time: Barycenter TESS Julian Day')
    else:
        logprint('Units of time: Julian Day')
    if mask ==1 :
        logprint('Mask: yes')
    else:
        logprint('Mask: no')

    logprint('Threshold limit for SNR: ', SNR_min)
    logprint('Threshold limit for SDE: ', SDE_min)
    logprint('Threshold limit for FAP: ', FAP_max)

    # We save here in a log file all the parameters available of the star.

    mass_min = mass-mass_min
    mass_max = mass+mass_max
    radius_min = radius-radius_min
    radius_max = radius+radius_max

    logprint('\n STELLAR PROPERTIES FOR THE SIGNAL SEARCH')
    logprint('================================================\n')
    logprint('limb-darkening estimates using quadratic LD (a,b)=', ab)
    logprint('mass =', format(mass,'0.5f'))
    logprint('mass_min =', format(mass_min,'0.5f'))
    logprint('mass_max =', format(mass_max,'0.5f'))
    logprint('radius =', format(radius,'0.5f'))
    logprint('radius_min =', format(radius_min,'0.5f'))
    logprint('radius_max =', format(radius_max,'0.5f'))



    logprint('\n SECTORS INFO')
    logprint('================================================\n')
    logprint('Sectors :', lcf)

    lc=lcf.PDCSAP_FLUX.stitch().remove_nans() # remove of the nans
    flux=lc.flux
    flux_err=lc.flux_err
    if time_units == 0:
        time=lc.astropy_time.jd #transform from TESS julian date to jd
    elif time_units == 1:
        time=lc.time  #keep the TESS julian date
    lc_new=lk.LightCurve(time=time, flux=flux,flux_err=flux_err)
    flux_clean=lc_new.remove_outliers(sigma_lower=float('inf'), sigma_upper=3) #remove outliers over 3sigma

    if mask==1:
        #first mask (example of how to mask a range)
        time_in=1449.8
        time_out=1451.50
        mask1 = (flux_clean.time < time_in) | (flux_clean.time > time_out)
        flux_clean = flux_clean[mask1]
        time_b=1463.5
        mask2 = (flux_clean.time < time_b)
        flux_clean = flux_clean[mask2]
        logprint('** Masking the lightcurve **')
    else:
        None


    # -------------------------------------------------------------------------------------------------------------------------------------------------------------
    #      Start of the runs
    # -------------------------------------------------------------------------------------------------------------------------------------------------------------
    print("________________________________ run 1 ________________________________")
    id_run = 1
    time_i = flux_clean.time
    flatten_i = flux_clean.flux
    results_pdcsap, SNR, key, periods, durations, tos = analyse(det_met, time_i, flatten_i, n_tra, ab, mass, mass_min, mass_max, radius, radius_min, radius_max, id_run)

    while key == 1:
        id_run += 1
        print("________________________________ run", id_run, "________________________________")
        SNR=np.nan_to_num(SNR)
        a = np.nanargmax(SNR)  # check the maximum SDE

        # applying a new mask
        time_i, flatten_i = add_mask(time_i, flatten_i, results_pdcsap, periods, durations, SNR, tos, a)

        # analyzing the data with the new mask
        results_pdcsap, SNR, key, periods, durations, tos = analyse(det_met, time_i, flatten_i, n_tra, ab, mass, mass_min, mass_max, radius, radius_min, radius_max, id_run)

@contextmanager
def elapsed_timer():
    from timeit import default_timer
    start = default_timer()
    elapser = lambda: str(default_timer() - start)
    yield lambda: elapser()
    end = default_timer()
    elapser = lambda: str(end - start)


with elapsed_timer() as elapsed:
    run()
    print(elapsed())

# end



from ..source import *
'''
def test_multi_sectors():
    tic = 29987116 # for testing
    assert(len(multi_sectors('all', tic=tic)) is 1)
    assert(len(multi_sectors([1], tic=tic)) is 1)
    assert(len(multi_sectors([3,6,8], tic=tic)) is 0)
    try:
        multi_sectors(1, tic=tic)
    except TypeError:
        pass
    except Exception as e:
       self.fail('Unexpected exception raised:', e)
    else:
       self.fail('TypeError not raised')
'''



import os
from urllib.request import urlopen
from datetime import datetime
import math
from astropy.coordinates import SkyCoord
from astropy import units as u
from astroquery.mast import Tesscut
from astropy.io import fits
import numpy as np
import sys
import requests
from bs4 import BeautifulSoup

eleanorpath = os.path.join(os.path.expanduser('~'), '.eleanor')
if not os.path.exists(eleanorpath):
    try:
        os.mkdir(eleanorpath)
    except OSError:
        eleanorpath = os.path.dirname(__file__)


def hmsm_to_days(hour=0, min=0, sec=0, micro=0):
    days = sec + (micro / 1.e6)
    days = min + (days / 60.)
    days = hour + (days / 60.)
    return days / 24.


def date_to_jd(year, month, day):
    if month == 1 or month == 2:
        yearp = year - 1
        monthp = month + 12
    else:
        yearp = year
        monthp = month

    # this checks where we are in relation to October 15, 1582, the beginning
    # of the Gregorian calendar.
    if ((year < 1582) or (year == 1582 and month < 10) or
            (year == 1582 and month == 10 and day < 15)):
        # before start of Gregorian calendar
        B = 0
    else:
        # after start of Gregorian calendar
        A = math.trunc(yearp / 100.)
        B = 2 - A + math.trunc(A / 4.)

    if yearp < 0:
        C = math.trunc((365.25 * yearp) - 0.75)
    else:
        C = math.trunc(365.25 * yearp)

    D = math.trunc(30.6001 * (monthp + 1))

    # including leap second correction
    jd = B + C + D + day + 1720994.5 + 0.0008

    return jd


def listFD(url, ext=''):
    page = requests.get(url).text
    soup = BeautifulSoup(page, 'html.parser')
    return [url + node.get('href') for node in soup.find_all('a') if
            node.get('href').endswith(ext)]


__all__ = ['Update', 'update_all']


def update_all():
    sector = 1
    good = 1
    while good:
        try:
            Update(sector=sector)
        except AttributeError:
            good = 0
        sector += 1


class Update(object):
    def __init__(self, sector=None):

        if sector is None:
            print('Please pass a sector into eleanor.Update().')
            return

        self.sector = sector
        self.metadata_path = os.path.join(eleanorpath, 'metadata/s{0:04d}'.format(self.sector))
        lastfile = 'cbv_components_s{0:04d}_0004_0004.txt'.format(self.sector)

        # Checks to see if directory contains all necessary files first
        if os.path.isdir(self.metadata_path):
            if lastfile in os.listdir(self.metadata_path):
                print('This directory already exists!')
                return

        self.north_coords = SkyCoord('16:35:50.667 +63:54:39.87',
                                     unit=(u.hourangle, u.deg))
        self.south_coords = SkyCoord('04:35:50.330 -64:01:37.33',
                                     unit=(u.hourangle, u.deg))

        if self.sector < 14 or self.sector > 26:
            try:
                manifest = Tesscut.download_cutouts(self.south_coords, 31, sector=self.sector)
                success = 1
            except:
                print("This sector isn't available yet.")
                return
        else:
            try:
                manifest = Tesscut.download_cutouts(self.north_coords, 31, sector=self.sector)
                success = 1
            except:
                print("This sector isn't available yet.")
                return


        if success == 1:
            if os.path.isdir(self.metadata_path) == True:
                pass
            else:
                os.mkdir(self.metadata_path)

        # memmap=False as wokaround for https://github.com/afeinstein20/eleanor/issues/204
        self.cutout = fits.open(manifest['Local Path'][0], memmap=False)

        print('This is the first light curve you have made for this sector. '
              'Getting eleanor metadata products for '
              'Sector {0:2d}...'.format(self.sector))
        print('This will only take a minute, and only needs to be done once. '
              'Any other light curves you make in this sector will be faster.')
        self.get_target()
        print('Target Acquired')
        self.get_cadences()
        print('Cadences Calculated')
        self.get_quality()
        print('Quality Flags Assured')
        self.get_cbvs()
        print('CBVs Made')
        print('Success! Sector {:2d} now available.'.format(self.sector))
        self.cutout.close()
        os.remove(manifest['Local Path'][0])
        self.try_next_sector()

    def get_cbvs(self):
        if self.sector <= 6:
            year = 2018
        elif self.sector <= 20:
            year = 2019
        else:
            year = 2020

        url = 'https://archive.stsci.edu/missions/tess/ffi/s{0:04d}/{1}/'.format(self.sector, year)

        directs = []
        for file in listFD(url):
            directs.append(file)
        directs = np.sort(directs)[1::]

        subdirects = []
        for file in listFD(directs[0]):
            subdirects.append(file)
        subdirects = np.sort(subdirects)[1:-4]

        for i in range(len(subdirects)):
            file = listFD(subdirects[i], ext='_cbv.fits')[0]
            os.system('curl -O -L {}'.format(file))

        time = self.cutout[1].data['TIME'] - self.cutout[1].data['TIMECORR']

        files = os.listdir('.')
        files = [i for i in files if i.endswith('_cbv.fits') and
                 's{0:04d}'.format(self.sector) in i]

        for c in range(len(files)):
            # memmap=False as wokaround for https://github.com/afeinstein20/eleanor/issues/204
            cbv      = fits.open(files[c], memmap=False)
            camera   = cbv[1].header['CAMERA']
            ccd      = cbv[1].header['CCD']
            cbv_time = cbv[1].data['Time']

            new_fn = eleanorpath + '/metadata/s{0:04d}/cbv_components_s{0:04d}_{1:04d}_{2:04d}.txt'.format(self.sector, camera, ccd)

            convolved = np.zeros((len(time), 16))
            for i in range(len(time)):
                g = np.argmin(np.abs(time[i] - cbv_time))
                for j in range(16):
                    index = 'VECTOR_{0}'.format(j+1)
                    if self.sector < 27:
                        cads = np.arange(g-7, g+8, 1)
                    else:
                        # XXX: need to test when TESSCut becomes available
                        cads = np.arange(g-2, g+3, 1)
                    convolved[i, j] = np.mean(cbv[1].data[index][cads])
            np.savetxt(new_fn, convolved)
            cbv.close()
        files = [i for i in files if i.endswith('_cbv.fits') and
                 's{0:04d}'.format(self.sector) in i]
        for c in range(len(files)):
            os.remove(files[c])

    def try_next_sector(self):
        codepath = os.path.dirname(__file__)
        f1 = open(codepath + '/maxsector.py', 'r')
        oldmax = float(f1.readline().split('=')[-1])
        if self.sector > oldmax:
            f = open(codepath + '/maxsector.py', 'w')
            f.write('maxsector = {:2d}'.format(self.sector))
            f.close()

    def get_target(self):
        filelist = urlopen('https://archive.stsci.edu/missions/tess/download_scripts/sector/tesscurl_sector_{:d}_lc.sh'.
                           format(self.sector))
        for line in filelist:
            if len(str(line)) > 30:
                import shutil
                os.system(str(line)[2:-3])
                fn = str(line)[2:-3].split()[5]
                shutil.move(fn, eleanorpath + '/metadata/s{0:04d}/target_s{0:04d}.fits'.format(self.sector))
                break
        return

    def get_cadences(self):
        if self.sector < 27:
            # these come from the first FFI cadence of S7, in particular
            # Camera 1, CCD 1 for the t0. The t0s vary by ~1 minute because of
            # barycentric corrections on different cameras
            index_zeropoint = 12680
            index_t0 = 1491.625533688852
        else:
            # first FFI cadence of S27 from Cam 1, CCD 1
            index_zeropoint = 116470
            index_t0 = 2036.283350837239

        times = np.array([], dtype=int)
        filelist = urlopen('https://archive.stsci.edu/missions/tess/download_scripts/sector/tesscurl_sector_{:d}_ffic.sh'.
                           format(self.sector))
        for line in filelist:
            if len(str(line)) > 30:
                times = np.append(times, int(str(line).split('tess')[1][0:13]))

        times = np.sort(np.unique(times))

        outarr = np.zeros_like(times)
        for i in range(len(times)):
            date = datetime.strptime(str(times[i]), '%Y%j%H%M%S')
            days = date.day + hmsm_to_days(date.hour, date.minute,
                                           date.second, date.microsecond)
            tjd = date_to_jd(date.year, date.month, days) - 2457000
            if self.sector < 27:
                cad = (tjd - index_t0)/(30./1440.)
            else:
                cad = (tjd - index_t0)/(10./1440.)
            outarr[i] = (int(np.round(cad))+index_zeropoint)

        np.savetxt(eleanorpath + '/metadata/s{0:04d}/cadences_s{0:04d}.txt'.format(self.sector), outarr, fmt='%i')
        return

    def get_quality(self):
        """ Uses the quality flags in a 2-minute target to create quality flags
            in the postcards.
        """

        ffi_time = self.cutout[1].data['TIME'] - self.cutout[1].data['TIMECORR']

        shortCad_fn = eleanorpath + '/metadata/s{0:04d}/target_s{0:04d}.fits'.format(self.sector)

        # Binary string for values which apply to the FFIs
        if self.sector > 26:
            ffi_apply = int('100000000010101111', 2)
        else:
            ffi_apply = int('100010101111', 2)

        # Obtains information for 2-minute target
        twoMin     = fits.open(shortCad_fn)
        twoMinTime = twoMin[1].data['TIME']-twoMin[1].data['TIMECORR']
        finite     = np.isfinite(twoMinTime)
        twoMinQual = twoMin[1].data['QUALITY']

        twoMinTime = twoMinTime[finite]
        twoMinQual = twoMinQual[finite]

        convolve_ffi = []
        for i in range(len(ffi_time)):
            where = np.where(np.abs(ffi_time[i] - twoMinTime) == np.min(np.abs(ffi_time[i] - twoMinTime)))[0][0]

            sflux = np.sum(self.cutout[1].data['FLUX'][i])
            nodata = 0
            if sflux == 0:
                nodata = 131072

            if (ffi_time[i] > 1420) and (ffi_time[i] < 1424):
                nodata = 131072

            if self.sector < 27:
                v = np.bitwise_or.reduce(twoMinQual[where-7:where+8])
            else:
                # XXX: need to test when TESSCut is available in S27
                v = np.bitwise_or.reduce(twoMinQual[where - 2:where + 3])
            convolve_ffi.append(np.bitwise_or(v, nodata))

        convolve_ffi = np.array(convolve_ffi)

        flags = np.bitwise_and(convolve_ffi, ffi_apply)

        np.savetxt(eleanorpath + '/metadata/s{0:04d}/quality_s{0:04d}.txt'.format(self.sector), flags, fmt='%i')



import os
import ast
import csv
import sys
import warnings
import matplotlib
import numpy as np
import pandas as pd
import tkinter as tk
from glob import glob

from os.path import exists
from tkinter import simpledialog
from argparse import ArgumentParser

#custom modules to import
from LATTE import LATTEbrew as brew 
from LATTE import LATTEutils as utils
#sys.tracebacklimit = 0
warnings.filterwarnings('ignore')

'''
NOTE: a warning message currently appears when the code is exited (i.e. at the end of the code). 
This is due to an error in the current verion of matplolib with the interactive widgets but has been adressed
in future releases of matplolib (see https://github.com/matplotlib/matplotlib/issues/13660)
The above link shows how to change the matplolib cbook __init__ file in order to make it disapear.

# REQUIRES MATPLOLIB 3.2 (there is a bug in 3.1 the current stable version - as of January 2020)). 
 --- pip install matplotlib==3.2.0rc1 (still in testing?)


Overview of LATTE scipts:

__main__.py	  : Intitialises the parameters, what TIC ID, sector, checks for downloads of the data, FFI or not? 

LATTEutils.py	: All the functions needed to download data and text files, runs the interactive gui, all of the plotting and data handling. 

LATTE_DV.py	  : Scipt to combine all of the results from LATTEbrew in order to generate a pdf data validation report.

LATTEbrew.py	 : Calls the LATTEutils functions in turn in order to store the results and keeps track of what has been generated. Calls the LATTE_DV.py function to collate all the results.

'''


# -------------------- START ---------------------
# ------------------------------------------------

if __name__ == '__main__':
	# these are all the input parameters that the user can (but doesn't have to) chose to identify in the command line.
	ap = ArgumentParser(description='Lightcurve Analysis Tool for Transiting Exoplanets')
	ap.add_argument('--new-data', action='store_true')
	ap.add_argument('--tic', type=str, help='the target TIC id, if blank will be prompted to enter', default = 'no')
	ap.add_argument('--sector', type=str, help='the sector(s) to look at', default = 'no')
	ap.add_argument('--targetlist', type=str, help='the link to the target file list', default='no')
	ap.add_argument('--noshow', action='store_true', help='if you want to NOT show the plots write --noshow in the command line')
	ap.add_argument('--o', action='store_true', help='if you call this old files will be overwriten in the non-interactive version')
	ap.add_argument('--auto', action='store_true', help='automatic aperture selection')
	ap.add_argument('--nickname', type=str, help='give the target a memorable name', default='noname')
	ap.add_argument('--FFI', action='store_true', help='is this an FIIs?')
	ap.add_argument('--save', help='is this an FIIs?', default = True)
	ap.add_argument('--north', action='store_true', help='write "north" if you want all the images to be aligned North')
	ap.add_argument('--new-path', action = 'store_true', help ='enter to change the output path')
	ap.add_argument('--mpi', action = 'store_true', help ='enter if the program is run with mpi')

	args = ap.parse_args()


	# ------------------------------------------------
	# Check what the current path is - when the program is first downloaded the path is set to 'no/path/set/yet' and the user is automatically prompted to change'no/path/set/yet'
	# Get the path to where the package is installed. This is where the configuration file and the images (for the DV report are stored)

	syspath = str(os.path.abspath(utils.__file__))[0:-14]
	# ----------

	def yes_or_no():
		'''
		Yes/No command line option to verify that the user really wants to change the output/input path
		'''
		print ('\n \n WARNING: if you have already downloded the input files (with --new-data) then these will remain in the location set by your previous path, so you will have to redowload the data (not recommended) or move the data to the new location set by this path. \n \n ')
	
		reply = str(input('Are you sure that you want to change the path?' + '(yes/no): '))
	
		if (reply == 'y') or (reply == 'yes') or (reply == 'yep') or (reply == 'yeah'):
			return True
	
		else: # if anything else is entered assume that this is a 'no' and continue with the old path
			return False	 
	
	# ------------------------------
	# ---- INITIALISE THE CODE -----
	# ------------------------------

	'''
	This first part of the code 1) ensures that an input and output path has been defined (and allows the user to change it if necessary)
	The path links to where the curl scipt files are stored and to where the output files will be stores. 2) Creates the output file if it doens't
	already exist, and 3) downloads the text files if run for the first time or if stated to download. 
	'''

	# check whether a path already exists
	if not os.path.exists("{}/_config.txt".format(syspath)):

		# if it doesn't exist ask the user to put it in the command line
		indir = input("\n \n No output path has been set yet. \n \n Please enter a path to save the files (e.g. ./LATTE_output or /Users/yourname/Desktop/LATTE_output) : " )

		#try three times to find the output path
		intry = 0
		worked = True

		while (intry < 3 ) and worked:
			if not os.path.exists("{}".format(indir)):
				indir = input("\n \n This path does not exist on your computer, please enter a valid  path: " )
	
				intry += 1

			else:
				worked = False
	
		if intry == 3:
			print ("You have entered an invalid path 3 times. \n Make sure that the path exists on your computer and that you have access to it before trying again.")
			sys.exit('')

		# SAVE the new output path
		with open("{}/_config.txt".format(syspath),'w') as f:
			f.write(str(indir))
		
		print("\n New path: " + indir)
	
		# this is also the first time that the program is being run, so download all the data that is required.
		print ("\n Download the text files required ... " )
		print ("\n Only the manifest text files (~325 M) will be downloaded and no TESS data." )
		print ("\n This step may take a while but luckily it only has to run once... \n" )

		# ------------------------------------------------
		#check whether the chosen (output) directory already exists, and if it doesn't create the directory.
		
		if not os.path.exists("{}/data".format(indir)):
			os.makedirs("{}/data".format(indir))
	
		# ------------------------------------------------

		# ----- REFERENCE FILES DOWNLOAD -----
		utils.data_files(indir)
		utils.tp_files(indir)
		utils.TOI_TCE_files(indir)
		utils.momentum_dumps_info(indir)
		
		# -----

	# if the user chooses to redefine the path
	elif args.new_path == True: 
	
		reply = yes_or_no() # double check that they really want to do that.
	
		if reply == True:
			indir = input("\n \n Please enter a path to save the files (e.g. ./LATTE_output or /Users/yourname/Desktop/LATTE_output) : " )

			#try three times to find the output path
			intry = 0
			worked = True
	
			while (intry < 3 ) and worked:
				if not os.path.exists("{}".format(indir)):
					indir = input("\n \n This path does not exist on your computer, please enter a valid  path: " )
		
					intry += 1
	
				else:
					worked = False
		
			if intry == 3:
				print ("You have entered an invalid path 3 times. \n Make sure that the path exists on your computer and that you have access to it before trying again.")
				sys.exit('')
	
			# SAVE the new output path
			with open("{}/_config.txt".format(syspath),'w') as f:
				f.write(str(indir))	
			
			print("\n New path: " + indir)
	
			if not os.path.exists("{}/data".format(indir)):
				os.makedirs("{}/data".format(indir))


		else:
			with open("{}/_config.txt".format(syspath), 'r') as f:
				indir = str(f.readlines()[-1])
				
			print ("LATTE will continue to run with the old path: {}".format(indir))

			#try three times to find the output path
			intry = 0
			worked = True
	
			while (intry < 3 ) and worked:
				if not os.path.exists("{}".format(indir)):
					indir = input("\n \n This old path does not exist on your computer, please enter a valid  path: " )
		
					intry += 1
	
				else:
					worked = False
		
			if intry == 3:
				print ("You have entered an invalid path 3 times. \n Make sure that the path exists on your computer and that you have access to it before trying again.")
				sys.exit('')
	
			if intry > 0:
				# SAVE the new output path
				with open("{}/_config.txt".format(syspath),'w') as f:
					f.write(str(indir))	

			if not os.path.exists("{}/data".format(indir)):
				os.makedirs("{}/data".format(indir))
				

	else:
		with open("{}/_config.txt".format(syspath), 'r') as f:
			indir = str(f.readlines()[-1])
	

		#try three times to find the output path
		intry = 0
		worked = True
	
		while (intry < 3 ) and worked:
			if not os.path.exists("{}".format(indir)):
				indir = input("\n \n Please enter a path to save the files (e.g. ./LATTE_output or /Users/yourname/Desktop/LATTE_output) : " )
		
				intry += 1
	
			else:
				worked = False
		
		if intry == 3:
			print ("You have entered an invalid path 3 times. \n Make sure that the path exists on your computer and that you have access to it before trying again.")
			sys.exit('')
	
		if intry > 0:
			# SAVE the new output path
			with open("{}/_config.txt".format(syspath),'w') as f:
				f.write(str(indir))	


		if not os.path.exists("{}/data".format(indir)):
			os.makedirs("{}/data".format(indir))
	
		# ------------------------------------------------

	'''
	Check whether to download the data reference files
	This will only run if you tell the program to run this (with the args)- needs to be run the first time that one wants to download data
	This should also be called if new TESS data is released 
	The program will check what data has already been downloaded and only download new data files.
	'''

	if (args.new_data != False) and (os.path.exists("{}/_config.txt".format(syspath))): 

		# ----- REFERENCE FILES DOWNLOAD -----
		utils.data_files(indir)
		utils.tp_files(indir)
		utils.TOI_TCE_files(indir)
		utils.momentum_dumps_info(indir)
		# -----
	


	# -----------  INTERACTIVE VERSION  ------------
	# ---------------------------------------------
	
	'''
	This section starts the interactive versio of the code:

	1) asks to enter the TIC ID
	2) identifies in which sectors the target was observed and asks the user to identify which sectors should be analysed

	NOTE: this part of the code requires you to have Tkinter installed. Tkinter currently does not work with certain new Mac operating systems.
	In order to run LATTE with an input list and not interatcively, state the path to the csv file when running the program.
	csv file must have format: "TICID, sectors, transits, BLS, model" - see example.
	The TIC ID and Sectors to look at can also be stated in the command line as an argument to save time
	The interactive tool is for ease of use and to avoid having to understand how to use the command line.
	'''
	
	# Check whether the a target list has been defined in the command line. If so, the code will not ask for a TIC ID or a sector 
	# as these would be listed in the input file and the code will run automatically. 
	if args.targetlist == 'no': 

		# Check whether the tic ID and the sector have already been entered
		# If both the sectors and the TIC ID are already entered then TKinter does not need to be loaded
		if args.tic != 'no' and args.sector != 'no':
			tic = str(args.tic)
			sectors = str(args.sector)

			# Check whether we are looking at an FFI
			# The FFI information needs to be stored straight away so a folder needs to be created to store them. 
			# The folder for the non-FFIs is created later after the user choses whether to 'save' data or not.
			if args.FFI == True:
				newpath = '{}/{}'.format(indir,tic)

				# if this folder doesn't exist then create it...
				if not exists(newpath):
					os.makedirs(newpath)

			# --------
			# Run a function called TESS-point. This returns the sectors in which the target has 
			# been observed as well as the RA and DEC of the target. 
			sectors_all, ra, dec = utils.tess_point(indir, tic) 
			# --------

		# --------------------------------
		# If either the TIC or the sectors or both have not already been identified, run Tkinter (interactive tools)
		
		else:

			# make a GUI interface with TKinter
			ROOT = tk.Tk()
			ROOT.withdraw()

			# if this hasn't already been identified as an FFI through the command line, give the option to chose this when entering the TIC ID
			if args.FFI == False:
				# -----------
				class TICprompt(simpledialog.Dialog):
				
					def body(self, master):
					
						# make a text box for the TIC ID
						tk.Label(master, text="Enter TIC ID:").grid(row=0)
						self.e1 = tk.Entry(master)
						self.e1.grid(row=0, column=1)
						
						# make a check button with the option to run this in FFI mode
						self.FFIbox = tk.IntVar()
						self.answer = tk.Checkbutton(master,text="Check for FFI mode", variable=self.FFIbox)
						self.answer.grid(row=1, column=1,  columnspan=2)
				
					def apply(self):
						# make these global variables so that they can be used outside of this class and applied to the rest of the program
						global tkTIC
						global tkFFI
						
						ROOT.form=(self.FFIbox.get())
						tkTIC = (self.e1.get())
						tkFFI =  (self.FFIbox.get())
				# -----------

				TICprompt(ROOT)
				# make the TIC a string
				tic = str(tkTIC)

				# If the FFI button was checked, change the FFI argument 

				if tkFFI == 1:
					args.FFI = True

			else: # if this is a FFI, don't give the FFI option again
				# has the TIC already been defined? If not, ask for the TIC ID with a text box. 
				if args.tic == 'no':
	
					# load first prompt window which asks for TIC ID	
					tic = simpledialog.askstring(title="TIC",
													  prompt="Enter TIC ID:")
				else:
					tic = str(args.tic)

			# -----------

			# Is this an FFI? - if it is, need to create a folder to store the FFI data in (one for each TIC ID)
			if args.FFI == True:
				newpath = '{}/{}'.format(indir,tic)
				# if this folder doesn't exist then create it...
				if not exists(newpath):
					os.makedirs(newpath)

			# has the sector already been defined? 
			if args.sector == 'no':
				# returns all of the sectors in which TESS observed the given TIC id - this uses TESS-point

				sectors_all, ra, dec = utils.tess_point(indir, tic) 
		
				# The user is shown a list of the sectors in which the target is observed 
				# and asked to state which ones they want to assess

				sectors = simpledialog.askstring(title="Sectors",
												  prompt="TIC {} was observed in sector(s):\n {}. \n Enter the ones you wish to look at (e.g. 1,4) or 'all' for all of them.".format(tic, sectors_all))
			else:
				# still need to run tess point even if the targets are already defined as we need to check whether target appears in the stated sector - sanity check
				sectors_all, ra, dec = utils.tess_point(indir, tic) 
				sectors = str(args.sector)

			# close the TKinter windows.
			ROOT.quit()
			ROOT.destroy()
		
		# if no sector is defined or the word 'all' is written in the box, analyse all of the given sectors.
		if len(sectors) == 0:
			sectors = 'all'

		
		# if not all of them are chosen, convert the input list (given as a string) into a python readable list
		if sectors != 'all':
			sectors = sectors.split(",")
			sectors = [int(i) for i in sectors]
		
		print ("\n")
		
		# print out the information that has been chosen to the command line.
		if sectors == 'all':
			print ("Will look at sector(s): {} \n".format(str(sectors_all)[1:-1]))
			sectors = sectors_all
		else:
			print ("Will look at sector(s):  {} \n".format(str(sectors)[1:-1]))
	

		# -------------------------------------
		# ---- OPEN INTERACTIVE MATPLOTLIB ----
		# -------------------------------------

		''' Start up LATTE interactive where the transit times can be chosen manually 
		 	this works differently for FFI data and target files as the data has a different format. 
		 	At this point the code continues in the LATTE_utils scipt. 
		 	This scipt (__main__.py) is only to set up the parameters for the target and to intitialise the code. 
		'''

		if args.FFI == False:
			utils.interact_LATTE(tic, indir, syspath, sectors_all, sectors, ra, dec, args)  # the argument of whether to shos the images or not 
		else:
			utils.interact_LATTE_FFI(tic, indir, syspath, sectors_all, sectors, ra, dec, args)
		
		# Done.

	# ---------------------------------------
	# ---------------------------------------
	#			RUN WITH INPUT FILE
	# ---------------------------------------
	#The below code is executed if the 'input target list' option has been chose - this is defined in the command line. 
	#The code is run with input targetlist - either with phase fold information or with transit time information.

	else:
		try:
			targetlist = pd.read_csv("{}".format(args.targetlist)) # If a path is defined, open the input file
		except:
			print ("This target list can't be found. Check the path you have given and the name and format of the file.")

		# check what kind of target list it is - with phase fold information or with period and t0 information
		if 'period' in targetlist.columns:
			pp = True # pp = phase fold
		else:
			pp = False

		# process each target individually
		for index, row in targetlist.iterrows():
			
			# ---- INPUT PARAMETERS ------
			try:
				tic = str(int(row['TICID']))
			except:
				continue

			# check whether this file already exist
			# if it already exists it will only be overwritten if --o function has been enabled to avoid file loss.
			existing_files = glob("{}/*{}*".format(indir, tic))
			
			if (len(existing_files) > 0)  and (args.o != True): 
				print ("This file already exists therefore SKIP. To overwrite files run this code with --o in the command line.")
				failed_tics = [] #keep track of the TIC IDs that failed to complete
				continue


			# --- WHAT SECTORS WAS IT OBSERVED IN? ---

			# load tess-point to identify the sectors tht it was observed in.
			# need to know whether the target actually was observed in the stated sector. 
			sectors_all, ra, dec = utils.tess_point(indir, tic)
			
			# convert the input list of sectors (string) to a list of numbers.
			try:
				sectors_in = str(int(row['sectors']))
				sectors_in = ast.literal_eval(sectors_in)
				if (type(sectors_in) == int) or (type(sectors_in) == float):
					
					sectors_in = [sectors_in]
				else:
					sectors_in = list(sectors_in)
				
				# Sucessfully entered sectors
				# check that the target was actually observed in the stated sector
				sectors = list(set(sectors_in) & set(sectors_all))

				if len(sectors) == 0:
					print ("The target was not observed in the sector(s) you stated ({}). \
							Therefore take all sectors that it was observed in: {}".format(sectors, sectors_all))
					sectors = sectors_all
			except:
				sectors = sectors_all

			failed_tics = [] #keep track of the TIC IDs that failed to complete


			# ---- IF NO TRANSIT MARKED RUN WITH INTERFACE ----
			if (pp == False) and (type(row['transits']) == float):
				utils.interact_LATTE(tic, indir, syspath, sectors_all, sectors, ra, dec, args)

			else:

				if pp == False: # if the transits were identified and no period information
					
					# extract the information of the tiems of the transit like events. 
					transit_list_in = (row['transits'])
					transit_list = ast.literal_eval(transit_list_in)
					
					# convert the input transit times and sectors into transit_list in the form of a list
					
					if (type(transit_list) == float) or (type(transit_list) == int):
						transit_list = [transit_list]
					else:
						transit_list = list(transit_list)
				
				# if the user entered a T0 adn a period, the code will calcualte the times if the transit like events and use that. 

				else:
					# get up to 3 markings - more than that will just be cluttered and will take too long - this can be changed later if more or less are desired. 
					transit_list = []
					period = row['period']
					t0 = row['t0']
					for i in range(0,4):
						transit = t0 + (i*period)
						if transit < (t0 + 20):
							transit_list.append(float(transit))

				# extract the other parameters from the input file.
				BLS_in = row['BLS']
				model_in = row['model']
				FFI_in = row['FFI']
	
				# ---- do you want to run a BLS? ----
				if (BLS_in == 'yes') or (BLS_in == 'True') or (BLS_in == 'true') or (BLS_in == '1'):
					BLS = True
				else:
					BLS = False
				
				# ---- do you want to model this transit? ----
				if (model_in == 'yes') or (model_in == 'True') or (model_in == 'true') or (model_in == '1'):
					model = True
				else:
					model = False
				
				# ---- run in FFI mode? ----
				if (FFI_in == 'yes') or (FFI_in == 'True') or (FFI_in == 'true') or (FFI_in == '1'):
					args.FFI = True


				# --- other SETTINGS ---
				simple = False  # we don't want to run the simple version - that's option is simply to do quick test
				save = True  # always save the files - no point running this if you're not going to save them
				DV = True   # we'll make a DV report for all of them
				args.noshow = True  # don't show these, just save them
				args.auto = True # the code will automatically choose it's own apertures in this mode

				if args.FFI == False:
					try:
						# ----------------------------------------
						# 			 DOWNLOAD DATA 
						# ----------------------------------------
						alltime, allflux, allflux_err, all_md, alltimebinned, allfluxbinned, allx1, allx2, ally1, ally2, alltime12, allfbkg, start_sec, end_sec, in_sec, tessmag, teff, srad = utils.download_data(indir, sectors, tic)
					
						# ----------------------------------------
						#			   START BREWING ....
						# ----------------------------------------
						
						brew.brew_LATTE(tic, indir, syspath,transit_list, simple, BLS, model, save, DV, sectors, sectors_all, alltime, allflux, allflux_err, all_md, alltimebinned, allfluxbinned, allx1, allx2, ally1, ally2, alltime12, allfbkg, start_sec, end_sec, in_sec, tessmag, teff, srad, ra, dec, args)
					except:

						failed_tics.append(tic)
						print ("TIC {} failed to complete. Continue anyway. You can find a list of the failed IDs stored in the output folder.".format(tic))
						continue
				else:
					try:
						# ----------------------------------------
						# 			 DOWNLOAD DATA 
						# ----------------------------------------

						alltime0, allflux_list, allflux_small, allflux0, all_md, allfbkg, allfbkg_t,start_sec, end_sec, in_sec, X1_list, X4_list, apmask_list, arrshape_list, tpf_filt_list, t_list, bkg_list, tpf_list = utils.download_data_FFI(indir, sectors, syspath, sectors_all, tic, args)
	
						# ----------------------------------------
						#			   START BREWING ....
						# ----------------------------------------
						brew.brew_LATTE_FFI(tic, indir, syspath, transit_list, simple, BLS, model, save, DV, sectors, sectors_all, alltime0, allflux_list, allflux_small, allflux0, all_md, allfbkg, allfbkg_t, start_sec, end_sec, in_sec, X1_list, X4_list, apmask_list, arrshape_list, tpf_filt_list, t_list, bkg_list, tpf_list, ra, dec, args)
					except:
						failed_tics.append(tic)
						print ("TIC {} failed to complete. Continue anyway. You can find a list of the failed IDs stored in the output folder.".format(tic))
						continue	


		# ----- ALL FILES IN INFILE PROCESSED -----
		# save a list of the failed TIC IDs

		if len(failed_tics) > 0: # if any of the targets failed to complete
			# check if there is already a file with failed tic-ids - we don't want to overwrite it but create a new one
			failed_outpath = "{}/failed_files".format(indir)
		
			# --------
			if not os.path.exists(failed_outpath): # if this folder doesn't already exist, make it
				os.makedirs(failed_outpath)
			# --------
		
			failedfiles = np.sort(glob("{}/failed_tics*".format(failed_outpath)))
			
			if len(failedfiles) > 0:
				last_number = int((failedfiles[-1].split('_')[-1][0:-4]))
			else:
				last_number = 1
		
			with open('{}/failed_tics_{}.csv'.format(failed_outpath,str(last_number+1)), "w") as f:
				# save
				writer = csv.writer(f)
				for val in failed_tics:
						writer.writerow([val])


	# ---------------------------------------------------------------
	# ------------------------- LAST STEP ---------------------------
	# ---------------------------------------------------------------

	# end by changing the name of the folder to include the nicknane if defined
	# this allows users to keep track of their targets more easily e.g. Planet Hunters TESS candidates are named after pastries.
		
	if not args.nickname == 'noname':
		os.system("mv {}/{} {}/{}_{}".format(indir, tic, indir, tic, args.nickname))

	print ("\n  Complete! \n ")

# End.

